// (c) Jean MICHEL June 1993 -- any modifications must be signaled to the
// author

// cell.cpp: classes to compute life generations. This is completely
// independant from windows, and can be used on any system

#include <stdio.h>
#include "cell.h"
#include <fstream.h>
#include <strstream.h>
#include <stdlib.h>
#include <string.h>
#include "efns.h"
static cell INFINI(0x7ffdu,0x7falu),
	    MINFINI(-HALFBASE,-HALFBASE);
#define MAXCELLS 16000
static cellpop temp(new cell[MAXCELLS]);
BOOL B36S23=FALSE;

cellpop::cellpop(cell *vec){v=vec;}

static BOOL addtemp(cell c)
{ if(temp.pop>=MAXCELLS)
  { efns(WARNING_,"%d: too many cells",MAXCELLS); return FALSE; }
  temp.v[temp.pop++]=c;
  return TRUE;
}

void cellpop::copy(const cellpop& w)
{ pop=w.pop; if(v)delete v;v=new cell[pop+1];
  for(int i=0;i<pop;i++)v[i]=w.v[i];
  v[i]=INFINI;
}

void cellpop::clear_pop(){temp.pop=0;copy(temp);}

cellpop::cellpop(int lg)
{ pop=lg; v=new cell[pop+1];
  for(int i=0;i<pop;i++)v[i]=cell(i,0);
  v[i]=INFINI;
}

BOOL cellpop::operator==(cellpop& w)
{ if(pop!=w.pop)return FALSE;
  for(int i=0;i<pop;i++)if(v[i]!=w.v[i])return FALSE;
  return TRUE;
}

BOOL cellpop::in(cell c)
{ for(int i=0;i<pop;i++)if(v[i]==c)return TRUE;
  return FALSE;
}

static BOOL rule[512];

void initrule()
{ for(int i=0;i<512;i++)
  { int j=i, sum=0;
    for(;j;j>>=1)sum+=j&1;
    rule[i]=sum==3 ||(sum==4 && (i&16)) || (B36S23 && sum==6 && (i&16));
  }
}

void cellpop::nextgen(cellpop& w)// v=nextgen from w
{ cell *ind[9];                  // indexes in w for neighb
  static cell neighb[9];         // neighbours
  for(int i=0;i<9;i++){neighb[i]=MINFINI;ind[i]=w.v;}
  static cell d[9]={cell(-1,-1),cell(0,-1),cell(1,-1),cell(-1,0),cell(0,0),
		    cell(1,0),cell(-1,1),cell(0,1),cell(1,1)};
		   // relative position of neighbours
  static int code[9]={1,2,4,8,16,32,64,128,256};
  cell lim=INFINI, x;temp.pop=0;lim+=d[0];
  for(cell z=MINFINI/* previous cell */;z<lim;z=x)
  { cell *n=neighb;int sum;
    x=INFINI; // will become current cell
    for (i=0;i<9;i++,n++)
    { if(*n==z){*n=*ind[i]++;*n+=d[i];}
      if(*n<x){x=*n;sum=code[i];}
      else if(*n==x)sum+=code[i];
    }
    if(rule[sum] && !addtemp(cell(x))){pop=0;delete[]v;v=NULL;return;}
  }
  copy(temp);
}

static int compar(const void *a,const void *b)
{ unsigned long x=*(unsigned long *)a,y=*(unsigned long *)b;
  return x<y?-1:(x>y?1:0);}

void cellpop::transform(char orientation)
{ int dirx=orientation&0x1?-1:1,diry=orientation&0x2?-1:1;
  BOOL rot=orientation&0x4;
  for(int i=0;i<pop;i++)
  v[i]=cell(dirx*(rot?v[i].y():v[i].x()),diry*(rot?v[i].x():v[i].y()));
  qsort(v,pop,sizeof(cell),compar);
}

void cellpop::add(cellpop&w,cellpop &shape,cell offset,int mode)
{ int iw=0;temp.pop=0;
  for(int is=0;is<shape.pop;is++)
  { cell c=shape.v[is];c+=offset;
    while(w.v[iw]<c)addtemp(w.v[iw++]);
    if(w.v[iw]>c){if(mode!=LAY_OFF)addtemp(c);}
    else if(mode==LAY_ON)addtemp(w.v[iw++]);
    else iw++;
  }
  while(iw<w.pop)addtemp(w.v[iw++]);
  copy(temp);
}

cell cellpop::lastx(void)
{ cell res;int x=-INFINI.x();
  for(int i=0;i<pop;i++)if(v[i].x()>=x){x=v[i].x();res=v[i];}
  return res;
}

cell cellpop::firstx(void)
{ cell res;int x=INFINI.x();
  for(int i=0;i<pop;i++)if(v[i].x()<x){x=v[i].x();res=v[i];}
  return res;
}

cell cellpop::firsty(void)
{ cell res;int y=INFINI.x();
  for(int i=0;i<pop;i++)if(v[i].y()<y){y=v[i].y();res=v[i];}
  return res;
}

cell cellpop::lasty(void)
{ cell res;int y=-INFINI.x();
  for(int i=0;i<pop;i++)if(v[i].y()>=y){y=v[i].y();res=v[i];}
  return res;
}

inline void swap(int &x,int &y){int temp=x;x=y;y=temp;}

void cellpop::erase_rect(cell start,cell end)
{ int y=0;
  unsigned minx=start.c.x,miny=start.c.y,maxx=end.c.x,maxy=end.c.y;
  if(minx>maxx)swap(minx,maxx);
  if(miny>maxy)swap(miny,maxy);
  for(int x=0;x<pop;x++)
     if(v[x].c.x<minx || v[x].c.x>=maxx || v[x].c.y<miny || v[x].c.y>=maxy)
	v[y++]=v[x];
  v[y]=INFINI;
  pop=y;
}

cellpop cellpop::get_rect(cell start,cell end)
{ temp.pop=0;
  unsigned minx=start.x(),miny=start.y(),maxx=end.x(),maxy=end.y();
  if(minx>maxx)swap(minx,maxx);
  if(miny>maxy)swap(miny,maxy);
  cell middle((minx+maxx)/2,(miny+maxy)/2);
  for(int i=0;i<pop;i++)
    if(v[i].x()>=minx && v[i].x()<maxx && v[i].y()>=miny && v[i].y()<maxy)
    { cell c=v[i];c-=middle;addtemp(c);}
  return cellpop(temp);
}

void cellpop::save(char *FileName)
{ ofstream os(FileName);
  if(!os){ efns(WARNING_|FILE_,FileName);return;}
  int start=firstx().c.x,x=start,y=firsty().c.y;
  for(int i=0;i<pop;i++)
  { for(;y<v[i].c.y;y++){os<<"\n";x=start;}
    for(;x<v[i].c.x;x++)os<<" ";
    os<<"o";x++;
  }
  os<<"\n";
}

static int readtemp(istream& is)
{ temp.pop=0;
  int xbase=0,x=xbase,y=0,c;
  while(strchr("\n\r .$\toO*x|#",c=is.get()))
  switch(c)
  { case '\n':y++;x=xbase;break;
    case '\r':break;
    case '#': if(is.get()=='P'){if(is>>xbase)is>>y;x=xbase;}
	      // fall through;
    case '|': while(EOF!=(c=is.get()) && c!='\n');break;
    case ' ':case '.':x++;break;
    case '$':x+=10;break;
    case '\t': x+=8-x%8;break;
    case 'o':case 'O': case '*':case 'x':
	     if(!addtemp(cell(x,y)))
	     { efns(WARNING_,"error occured at line %d",y);
	       return -y;
	     }
	     x++;break;
  }
  if(c!=EOF)is.putback(c);
  qsort(temp.v,temp.pop,sizeof(cell),compar);
  return y;
}

cellpop::cellpop(char *string)
{ istrstream is(string);
  int line=readtemp(is);
  if(EOF!=is.get())
  { char buf[15];is.get(buf,15);
    efns(WARNING_,"error in input <%.15s> at line %d",buf,line);
  }
  v=0;copy(temp);
}

cellpop readfile(char *FileName)
{ ifstream is(FileName);
  if(!is){ efns(WARNING_|FILE_,FileName);return cellpop();}
  int line=readtemp(is);
  if(EOF!=is.get())efns(WARNING_,"error reading %s at line %d",FileName,line);
  return temp;
}

struct named_shape
{ char *name;
  cellpop shape;
  named_shape *next,*prev;
};

static named_shape *shapeptr=NULL;

static int shapeidx=-1;

char *getshapename(int i)
{ if(!shapeptr)return NULL;
  while(i>shapeidx && shapeptr->next)
  { shapeidx++;shapeptr=shapeptr->next;}
  if(i>shapeidx)return NULL;
  while(i<shapeidx && shapeptr->prev)
  { shapeidx--;shapeptr=shapeptr->prev;}
  if(i<shapeidx)return NULL;
  return shapeptr->name;
}

cellpop getshape(int i)
{ getshapename(i);
  return shapeptr->shape;
}

#define MAXNAME 30
void readshapelib(char *FileName)
{ char name[MAXNAME+1];
  named_shape *oldptr=NULL;
  ifstream is(FileName);
  if(!is){ efns(WARNING_|FILE_,FileName);return;}
  if(shapeptr)
  { while(shapeptr->next)
    { shapeidx++;shapeptr=shapeptr->next;}
    oldptr=shapeptr;
  }
  int c, line=1;
  while((c=is.get())!=EOF)switch(c)
  { case '|':while((c=is.get())!=EOF && c!='\n');line++;break;
    case '[':{for(char *n=name;(c=is.get())!=']';n++)
	     { if(n-name>=MAXNAME)
	       { efns(WARNING_,"error in %s line %d: name too long (>%d chars)",
		    FileName,line,MAXNAME);
		 goto fin;
	       }
	       if(c==EOF || c=='\n')
	       { efns(WARNING_,
		      "error in %s line %d: line break or end of file in name",
		    FileName,line);
		 goto fin;
	       }
	       *n=c;
	     }
	     *n=0;
	     }
	     while((c=is.get())==' ' || c=='\t' || c=='\r');
	     if(c!='\n')
	     { efns(WARNING_,"error in %s: garbage follows name at line %d",
		  FileName,line);
	       goto fin;
	     }
	     else line++;
	     line+=readtemp(is);
	     { cellpop w=temp;
	       named_shape *tmp=new named_shape;
	       if(shapeptr)shapeptr->next=tmp;
	       tmp->prev=shapeptr;
	       shapeptr=tmp;
	       shapeidx++;
	       shapeptr->next=NULL;
	       shapeptr->name=strdup(name);
	       shapeptr->shape=w;
	     }
	     break;
    default: efns(WARNING_,"error in %s line %d <%c>",FileName,line,c);
	     goto fin;
  }
  fin:
  if(oldptr)
  { oldptr->next->prev=NULL;
    for(;oldptr;oldptr=oldptr->prev)shapeidx--;
  }
  return;
}
