// (c) Jean MICHEL June 1993 -- any modifications must be signaled to the
// author

// cell.h: classes to compute life generations. This is completely
// independant from windows, and can be used on any system

typedef int BOOL;
#define FALSE 0
#define TRUE 1

typedef unsigned long INTCELL;
#define HALFBASE 0x8000u
class cell // point (x,y) coded as (x+HALFBASE)+0x10000lu*(y+HALFBASE)
	   // sorted, end marked by INFINI
{ public:
  union {struct{unsigned x,y;}c;
	 INTCELL i;};
  cell(){i=0;}
  cell(unsigned xx,unsigned yy){c.x=xx+HALFBASE;c.y=yy+HALFBASE;}
  void operator +=(const cell& d){i+=d.i+0x7fff8000lu;}
  void operator -=(const cell& d){i+=-d.i+0x80008000lu;}
  BOOL operator ==(cell c){return i==c.i;}
  BOOL operator !=(cell c){return i!=c.i;}
  BOOL operator <(cell c){return i<c.i;}
  BOOL operator <=(cell c){return i<=c.i;}
  BOOL operator >(cell c){return i>c.i;}
  int x(){return (int)(c.x-HALFBASE);}
  int y(){return (int)(c.y-HALFBASE);}
};

enum {UNSET,SET,RESET};
enum {LAY_ON=1,LAY_OFF=2,LAY_XOR=3};

// cellpop: a population of cells

class cellpop
{ public:
  cell *v;
  int pop;                // number of cells
  cellpop(){v=NULL;pop=0;}// default constructor (empty population)
  cellpop(cell *);   // constructor from a vector of cell
  cellpop(char *);        // constructor from a string
  cellpop(int n);         // constructs a line of n cells
  ~cellpop(){if(v)delete[] v;}
  void copy(const cellpop& w);
  cellpop(const cellpop& w){v=NULL;copy(w);}     // copy constructor
  void operator=(const cellpop& w){copy(w);}  // assignment operator
  BOOL operator==(cellpop& w);       // equality test
  void nextgen(cellpop& w);          // w is next generation from this
  BOOL in(cell c);                   // tests if c is in this
  BOOL error(){return v==NULL;}
  void erase_rect(cell start,cell end); // erase all cells between bounds
  void save(char *FileName);            // save population to file
  cellpop get_rect(cell start,cell end); // gets cells within bounds
  cell firstx(void);            // find leftmost cell
  cell firsty(void);            // find topmost cell
  cell lastx(void);             // find rightmost cell
  cell lasty(void);             // find bottommost cell
  void transform(char orientation);  // rotate pop. or mirror it
  void add(cellpop&v,cellpop &shape,cell offset,int mode);
				     // this=v+shape added at given offset
				     // where mode is one of ON, OFF,XOR
  void clear_pop();
};

cellpop readfile(char *filename);    // read pop. from file
void readshapelib(char *FileName);   // load shape library
cellpop getshape(int i);             // find ith shape in library
char *getshapename(int i);           // find name of ith shape in library
extern BOOL B36S23;
void initrule();
