// (c) Jean MICHEL June 1993, June 1995 --
// any modifications must be signaled to the author

#pragma hdrfile "owlall.csm"
#include <owl\owlall.h>
#pragma hdrstop
#include <stdio.h>
#include <stdlib.h>
#include <dir.h>
#include <io.h>
#include <string.h>
#include "efns.h"
#include "cell.h"
#include "shapewnd.h"
#include "life.h"
#include "life.rc"
TPen *Pens[3]; TBrush *Brushes[3]; TColor Colors[3];

TStatusBar *statusbar;

class TLife : public TApplication
{public:
  TLife() : TApplication(){};
  void InitMainWindow();
};

_CLASSDEF(TLifeWindow)
class TLifeWindow : public TWindow
{
protected:
  TDC *DrawDC;

public:
  cellpop v;
  cell box_start,box_end;
  TDC *boxDC;
  BOOL show_grid,boxpresent;
  int magnification;
  RECT ScrollRange;
  long gencounter;int showlimit;
  struct {long x;long y;}origin;// Logical coordinates of top left
  TShapeWindow  *ShapeWindow;
  void setorigin(), showgen(), showpop();
  void DrawBox(TDC *DC);
  void SetBox(BOOL state);
  void Run(long count);
  cell coordtocell(TPoint&);
  int update(cellpop& w);
  void Paint(TDC&,BOOL,TRect&);
  void pixel(TDC *DC,cell c,int color);
  TLifeWindow(TWindow *Parent=0);
  ~TLifeWindow();
  void adjustscroll(int newmag);
  virtual void SetupWindow();
  void EvLButtonDown( UINT, TPoint& );
  void EvRButtonDown( UINT, TPoint& );
  void EvRButtonUp( UINT, TPoint& );
  void EvMouseMove( UINT, TPoint& );
  void CMShape();
  void CMUFileSave();
  void CMUFileSaveArea();
  void CMErase();
  void CMEditCut();
  void CMEditCopy();
  void CMColorSet();
  void CMColorReset();
  void CMColorUnset();
  void CMClearGen();
  void CMShowGrid();
  void CMB36S23();
  void CMUColor(int index);
  void CMFileNew();
  void CMHelp();
  void CMHelphelp();
  void CMAbout();
  void CMZoomOut();
  void CMZoomIn();
  void CMStep();
  void CMStep2();
  void CMStep4();
  void CMRun();
  void CeBoxEnabler(TCommandEnabler& commandEnabler);
  virtual LPSTR GetClassName();
  void GetWindowClass(WNDCLASS& AWndClass);
DECLARE_RESPONSE_TABLE( TLifeWindow );
};

int mags[]={1,2,3,5,7,10,15,25};int magindex=4;

LPSTR TLifeWindow::GetClassName(){ return "LifeWindow";}

void TLifeWindow::GetWindowClass(WNDCLASS& AWndClass)
{ TWindow::GetWindowClass(AWndClass);
  AWndClass.hIcon=LoadIcon(AWndClass.hInstance,"LifeIcon");
}

TLifeWindow::TLifeWindow(TWindow *Parent)
{ Init(Parent,"",0);
  Attr.Style|= WS_VSCROLL|WS_HSCROLL;
  magnification=mags[magindex];origin.x=origin.y=0;
  boxpresent=show_grid=FALSE;
  boxDC=NULL;
  Scroller=new TScroller(this,1,1,1,1);
  // meaningless values -- setup by SetupWindow
  Colors[UNSET]=RGB(120,255,120);
  Colors[SET]=RGB(255,0,0);
  Colors[RESET]=RGB(200,0,0);
  for(int i=0;i<3;i++)
  { Brushes[i]=new TBrush(Colors[i]);
    Pens[i]=new TPen(Colors[i]);
  }
  ShapeWindow=new TShapeWindow(NULL);
}

void TLifeWindow::SetupWindow()
{ TWindow::SetupWindow();
  TRect r=GetClientRect();
  ScrollRange.left=-(r.right-r.left)/(2*magnification);
  ScrollRange.top=-(r.bottom-r.top)/(2*magnification);
  ScrollRange.right=ScrollRange.left+1;
  ScrollRange.bottom=ScrollRange.top+1;
  Scroller->AutoOrg=FALSE;
  v=cellpop();gencounter=0;showgen();update(v);
  showlimit=1;
  adjustscroll(magnification);
  readshapelib("life.lib");
  ShapeWindow->Create();
  selected_shape=cellpop("o\n");
  initrule();
}

TLifeWindow::~TLifeWindow()
{ for(int i=0;i<3;i++)
  { delete Brushes[i];
    delete Pens[i];
  }
}

void TLifeWindow::Paint(TDC &DC,BOOL Erase,TRect &r)
{ setorigin();
  DC.SelectObject(*Pens[RESET]);
  DC.SelectObject(*Brushes[RESET]);
  cell first=coordtocell(r.TopLeft()),last=coordtocell(r.BottomRight());
  for(int i=0;v.v[i]<=last;i++)if(first<=v.v[i])pixel(&DC,v.v[i],RESET);
  if(show_grid && magnification>3)
  { TPen *pen=new TPen(TColor(100,100,100));
    DC.SelectObject(*pen);
    for(int x=r.left;x<r.right;x+=magnification)
    { DC.MoveTo(x,r.top);DC.LineTo(x,r.bottom);}
    for(x=r.top;x<r.bottom;x+=magnification)
    { DC.MoveTo(r.left,x);DC.LineTo(r.right,x);}
    DC.RestorePen();
  }
  if(boxpresent)DrawBox(&DC);
}

void TLifeWindow::CeBoxEnabler(TCommandEnabler& cE)
{cE.Enable(boxpresent);}

void TLifeWindow::showgen()
{ char text[18];sprintf(text,"generation: %5d",gencounter);
  ((TTextGadget *)((*statusbar)[1]))->SetText(text);
}

void TLifeWindow::showpop()
{ char text[18];sprintf(text,"population: %5d",v.pop);
  statusbar->SetText(text);
}

int TLifeWindow::update(cellpop& w)
{ int nbchg=0;
  DrawDC=new TClientDC(*this);
  setorigin();
  DrawDC->SelectObject(*Pens[UNSET]);
  DrawDC->SelectObject(*Brushes[UNSET]);
  for(int iw=0,iv=0;iw!=w.pop || iv!=v.pop;)
    if (w.v[iw]<v.v[iv])iw++;
    else
    { if(w.v[iw]!=v.v[iv]){pixel(DrawDC,v.v[iv],UNSET);nbchg++;}
      iv++;
    }
  DrawDC->SelectObject(*Pens[SET]);
  DrawDC->SelectObject(*Brushes[SET]);
  for(iw=0,iv=0;iw!=w.pop || iv!=v.pop;)
    if (w.v[iw]<v.v[iv]){pixel(DrawDC,w.v[iw],SET);nbchg++;iw++;}
    else
    { if(w.v[iw]==v.v[iv])iw++;
      iv++;
    }
  DrawDC->SelectObject(*Pens[RESET]);
  DrawDC->SelectObject(*Brushes[RESET]);
  for(iw=0,iv=0;iw!=w.pop || iv!=v.pop;)
    if (w.v[iw]<v.v[iv])iw++;
    else
    { if(w.v[iw]==v.v[iv]){pixel(DrawDC,w.v[iw],RESET);iw++;}
      iv++;
    }
  delete DrawDC;
  v=w;
  showpop();
  return nbchg;
}

void TLifeWindow::setorigin()
{ origin.x=Scroller->XPos+ScrollRange.left;
  origin.y=Scroller->YPos+ScrollRange.top;
}

void TLifeWindow::pixel(TDC *DC,cell c,int color)
{ int x=magnification*(c.x()-origin.x);
  int y=magnification*(c.y()-origin.y);
  if(magnification>1)
  { int sz=magnification-(magnification>3);
    DC->Rectangle(x,y,x+sz,y+sz);
  }
  else DC->SetPixel(x,y,Colors[color]);
}

cell TLifeWindow::coordtocell(TPoint &p)
{ return cell(p.x/magnification+origin.x,p.y/magnification+origin.y);}

void TLifeWindow::EvLButtonDown(UINT, TPoint& p)
{ cellpop w=v;
  w.add(w,selected_shape,coordtocell(p),LAY_XOR);
  if(selected_shape.pop!=1)selected_shape="o";
  update(w);
}

void TLifeWindow::DrawBox(TDC *DC)
{ int x1=magnification*(box_start.x()-origin.x);
  int y1=magnification*(box_start.y()-origin.y);
  int x2=magnification*(box_end.x()-origin.x);
  int y2=magnification*(box_end.y()-origin.y);
  DC->SetROP2(R2_NOT);
  DC->MoveTo(x1,y1);
  DC->LineTo(x1,y2);
  DC->LineTo(x2,y2);
  DC->LineTo(x2,y1);
  DC->LineTo(x1,y1);
  DC->SetROP2(R2_COPYPEN);
}

void TLifeWindow::SetBox(BOOL state)
{ if(boxpresent!=state)
  { boxDC=new TClientDC(*this);
    DrawBox(boxDC);
    delete boxDC; boxDC=NULL;
    boxpresent=state;
  }
}

void TLifeWindow::EvRButtonDown( UINT, TPoint& p)
{ SetBox(FALSE);
  boxDC=new TClientDC(*this);
  box_start=box_end=coordtocell(p);DrawBox(boxDC);
  boxpresent=TRUE;
  SetCapture();
}

void TLifeWindow::EvMouseMove( UINT, TPoint& p)
{ if(boxDC){DrawBox(boxDC); box_end=coordtocell(p);DrawBox(boxDC);}
  char buf[20];
  sprintf(buf,"x=%5ld y=%5ld",p.x/magnification+origin.x,p.y/magnification+origin.y);
  ((TTextGadget *)((*statusbar)[3]))->SetText(buf);
}

void TLifeWindow::EvRButtonUp( UINT, TPoint& )
{ if(boxDC){delete boxDC;boxDC=NULL;ReleaseCapture();}
}

void TLifeWindow::Run(long count)
{ MSG msg;
  cellpop w,last2,last=v;int showcount=0;
  do
  {  w.nextgen(last);
     if(w.error())return;
     if(w.pop==0){efns(WARNING_,"dead");break;}
     if(w==last){efns(WARNING_,"stable");break;}
     if(w==last2){efns(WARNING_,"period 2");break;}
     last2=last;last=w;
     gencounter++;
     if(++showcount==showlimit){showgen();update(w);showcount=0;}
     if( PeekMessage(&msg,NULL,0,0,PM_REMOVE))
     { TranslateMessage(&msg);
       DispatchMessage(&msg);
       if((msg.message== WM_KEYDOWN)
	   ||(msg.message>= 0x201 && msg.message<=WM_MOUSELAST)
	   ||(msg.message>= 0xA3 && msg.message<=0xA9)
	   || msg.message== 0xA1)
	 break;
     }
  }
  while(--count>0);
  if(showcount<showlimit){showgen();update(w);showcount=0;}
}

void TLifeWindow::CMStep(){Run(showlimit=1);}

void TLifeWindow::CMStep2(){Run(showlimit=2);}

void TLifeWindow::CMStep4(){Run(showlimit=4);}

void TLifeWindow::CMRun() {Run(100000);}

void TLifeWindow::adjustscroll(int newmag)
{ RECT r;::GetClientRect(HWindow,&r);
  Scroller->XPos+=(newmag-magnification)*(r.right-r.left)/(2*newmag*magnification);
  Scroller->YPos+=(newmag-magnification)*(r.bottom-r.top)/(2*newmag*magnification);
  magnification=newmag;
  char text[18];sprintf(text,"magnification: %2d",magnification);
  ((TTextGadget *)((*statusbar)[2]))->SetText(text);
  Scroller->XLine=Scroller->YLine=25/magnification;
  Scroller->SetUnits(magnification,magnification);
  RECT NewRange;
  if(v.pop>0)
  { NewRange.left=v.firstx().x();
    NewRange.right=v.lastx().x();
    NewRange.top=v.firsty().y();
    NewRange.bottom=v.lasty().y();
  }
  else
  { NewRange.left=0;
    NewRange.right=0;
    NewRange.top=0;
    NewRange.bottom=0;
  }
  NewRange.left-=(r.right-r.left)/magnification;
  NewRange.top-=(r.bottom-r.top)/magnification;
  Scroller->XPos+=ScrollRange.left-NewRange.left;
  Scroller->YPos+=ScrollRange.top-NewRange.top;
  ScrollRange=NewRange;
  Scroller->SetRange(ScrollRange.right-ScrollRange.left,
		     ScrollRange.bottom-ScrollRange.top);
  Invalidate();
}

void TLifeWindow::CMZoomIn()
{ if(magnification<25)adjustscroll(mags[++magindex]);}

void TLifeWindow::CMZoomOut()
{ if(magnification>1)adjustscroll(mags[--magindex]);}

void TLifeWindow::CMFileNew()
{ v.clear_pop();showpop();CMClearGen();Invalidate();}

void TLifeWindow::CMEditCut(){CMEditCopy(); CMErase();}

void TLifeWindow::CMEditCopy()
{ SetBox(FALSE);
  selected_shape=v.get_rect(box_start,box_end);
  ShapeWindow->get();
}

void TLifeWindow::CMErase()
{ cellpop w=v;w.erase_rect(box_start,box_end);update(w);
  SetBox(FALSE);
}

void TLifeWindow::CMUFileSave()
{ TOpenSaveDialog::TData FileData(OFN_OVERWRITEPROMPT
  |OFN_HIDEREADONLY|OFN_PATHMUSTEXIST,
   "Life patterns (*.LFE)|*.lfe|All files (*.*)|*.*|",0,"","LFE");
//  strcpy(FileData.FileName,"*.lfe");
  if((TFileSaveDialog(this,FileData)).Execute()==IDOK)
    v.save(FileData.FileName);
}

void TLifeWindow::CMUFileSaveArea()
{ TOpenSaveDialog::TData FileData(OFN_OVERWRITEPROMPT,
   "Life patterns (*.lfe)|*.lfe|",0,0,"lfe");
  strcpy(FileData.FileName,"");
  if((TFileSaveDialog(this,FileData)).Execute()==IDOK)
  { SetBox(FALSE);
    v.get_rect(box_start,box_end).save(FileData.FileName);
  }
}

void TLifeWindow::CMColorUnset(){CMUColor(UNSET);}

void TLifeWindow::CMColorReset(){CMUColor(RESET);}

void TLifeWindow::CMColorSet(){CMUColor(SET);}

void TLifeWindow::CMClearGen(){gencounter=0;showgen();}

void TLifeWindow::CMShowGrid()
{ show_grid=!show_grid;
  CheckMenuItem(Parent->GetMenu(),CM_SHOW_GRID,MF_BYCOMMAND|
    (show_grid?MF_CHECKED:MF_UNCHECKED));
  Invalidate();
}

void TLifeWindow::CMB36S23()
{ B36S23=!B36S23;
  CheckMenuItem(Parent->GetMenu(),CM_B36S23,MF_BYCOMMAND|
    (B36S23?MF_CHECKED:MF_UNCHECKED));
}

void TLifeWindow::CMUColor(int index)
{ TChooseColorDialog::TData colors;
  static TColor custColors[16] =
  { 0x010101L, 0x101010L, 0x202020L, 0x303030L,
    0x404040L, 0x505050L, 0x606060L, 0x707070L,
    0x808080L, 0x909090L, 0xA0A0A0L, 0xB0B0B0L,
    0xC0C0C0L, 0xD0D0D0L, 0xE0E0E0L, 0xF0F0F0L
  };
  colors.Flags = CC_RGBINIT;
  colors.Color = Colors[index];
  colors.CustColors = custColors;
  if (TChooseColorDialog(this, colors).Execute() == IDOK)
  { Colors[index]=colors.Color;
    delete Brushes[index]; Brushes[index]=new TBrush(Colors[index]);
    delete Pens[index]; Pens[index]=new TPen(Colors[index]);
    Invalidate();
  }
}

void TLifeWindow::CMHelp(){WinHelp("LIFE.HLP",HELP_CONTENTS,0l);}

void TLifeWindow::CMHelphelp(){WinHelp(NULL,HELP_HELPONHELP,0l);}

void TLifeWindow::CMAbout()
{ TDialog(this, IDD_ABOUT).Execute();}

void TLifeWindow::CMShape()
{ ShapeWindow->display();}

void TLife::InitMainWindow()
{ TDecoratedFrame *frame=new TDecoratedFrame(NULL,
    "Windows Life Version 0.2",new TLifeWindow,TRUE);
  statusbar=new TStatusBar(0);
  statusbar->Insert(*new TTextGadget(TGadget::None,TGadget::Recessed,TTextGadget::Left,13));
  statusbar->Insert(*new TTextGadget(TGadget::None,TGadget::Recessed,TTextGadget::Left,13));
  statusbar->Insert(*new TTextGadget(TGadget::None,TGadget::Recessed,TTextGadget::Left,16));
  frame->AssignMenu("COMMANDS");
  frame->Attr.AccelTable="COMMANDS";
  frame->Insert(*statusbar,TDecoratedFrame::Bottom);
  frame->SetIcon(this,"LifeIcon");
  SetMainWindow(frame);
}

#include "winerror.h"
void (cdecl *efns)(int errcode,char *fmt,...)=error;

int OwlMain(int,char*[]){return TLife().Run();}

DEFINE_RESPONSE_TABLE1( TLifeWindow, TWindow )
    EV_WM_LBUTTONDOWN,
    EV_WM_RBUTTONDOWN,
    EV_WM_RBUTTONUP,
    EV_WM_MOUSEMOVE,
    EV_COMMAND( CM_SHAPE, CMShape ),
    EV_COMMAND( CM_U_FILESAVEAS, CMUFileSave ),
    EV_COMMAND( CM_U_FILESAVEAREAAS, CMUFileSaveArea ),
    EV_COMMAND( CM_ERASE, CMErase ),
    EV_COMMAND( CM_EDITCUT, CMEditCut ),
    EV_COMMAND( CM_EDITCOPY, CMEditCopy ),
    EV_COMMAND( CM_COLOR_SET, CMColorSet ),
    EV_COMMAND( CM_COLOR_RESET, CMColorReset ),
    EV_COMMAND( CM_COLOR_UNSET, CMColorUnset ),
    EV_COMMAND( CM_CLEAR_GEN, CMClearGen ),
    EV_COMMAND( CM_SHOW_GRID, CMShowGrid ),
    EV_COMMAND( CM_B36S23, CMB36S23 ),
    EV_COMMAND( CM_FILENEW, CMFileNew ),
    EV_COMMAND( CM_HELP, CMHelp ),
    EV_COMMAND( CM_HELPHELP, CMHelphelp ),
    EV_COMMAND( CM_ABOUT, CMAbout ),
    EV_COMMAND( CM_ZOOMOUT, CMZoomOut ),
    EV_COMMAND( CM_ZOOMIN, CMZoomIn ),
    EV_COMMAND( CM_STEP, CMStep ),
    EV_COMMAND( CM_STEP2, CMStep2 ),
    EV_COMMAND( CM_STEP4, CMStep4 ),
    EV_COMMAND( CM_RUN, CMRun ),
    EV_COMMAND_ENABLE( CM_EDITCUT, CeBoxEnabler),
    EV_COMMAND_ENABLE( CM_EDITCOPY, CeBoxEnabler),
    EV_COMMAND_ENABLE( CM_U_FILESAVEAREAAS, CeBoxEnabler),
    EV_COMMAND_ENABLE( CM_ERASE, CeBoxEnabler),
END_RESPONSE_TABLE;
