// (c) Jean MICHEL June 1993 -- any modifications must be signaled to the
// author

// class TShapeWindow is for the 'Shapes' window

#pragma hdrfile "owlall.csm"
#include <owl\owlall.h>
#pragma hdrstop
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "shapewnd.h"
#include "life.h"
#include "efns.h"
#define ID_LISTBOX  101
#define ID_BUTTON1  102
#define ID_BUTTON2  103
#define ID_BUTTON3  104
#define ID_BUTTON4  105
#define ID_BUTTON5  106
#define ID_BUTTON6  107
#define ID_BUTTON7  108

cellpop selected_shape("o");
int listbox_index;
const TRect TShapeWindow::r(20,165,220,365);

TShapeWindow::TShapeWindow(TWindow * AParent) :
   TWindow(AParent, "Select a shape")
{ Attr.Style = WS_POPUP|WS_BORDER|WS_CAPTION;
  Attr.X = 0;
  Attr.Y = 0;
  Attr.W = 240;
  Attr.H = 480;
  new TButton(this, ID_BUTTON7, "Load new shape library",   20, 10, 200, 25, FALSE);
  ListBox = new TListBox(this, ID_LISTBOX, 20, 45, 200, 75);
  ListBox->Attr.Style^=LBS_SORT;
  new TButton(this, ID_BUTTON3, "Load from file", 20, 115, 140, 25, FALSE);
  new TButton(this, ID_BUTTON4, "Line",  180, 115, 40, 25, FALSE);
  PreviewMsg=new TStatic(this, -1,  "Preview:", 20, 145, 160, 19, 0);
  new TButton(this, ID_BUTTON5, "Mirror",  30, 375, 80, 25, FALSE);
  new TButton(this, ID_BUTTON6, "Rotate",  130, 375, 80, 25, FALSE);
  new TButton(this, ID_BUTTON1, "Ok",      50, 410, 40, 30, TRUE);
  new TButton(this, ID_BUTTON2, "Cancel",  120, 410, 70, 30, TRUE);
}

void TShapeWindow::SetupWindow()
{ TWindow::SetupWindow();
  SetupList();
  show();
}

void TShapeWindow::SetupList()
{ char *n;
  ListBox->ClearList();
  for(int j=0;n=getshapename(j);j++)ListBox->AddString(n);
  ListBox->SetSelIndex(listbox_index);
  getshapefromlist();
}

void TShapeWindow::HandleListBoxMsg( UINT )
{ listbox_index=ListBox->GetSelIndex();
  getshapefromlist();
}

void TShapeWindow::getshapefromlist()
{ cellpop temp=cellpop(getshape(listbox_index));
  if(temp.error())return;
  local_shape=temp;
  selected_shape=local_shape;
  show();
}

void TShapeWindow::HandleButton2Msg( UINT )
{ selected_shape=cellpop("o\n");
  Show(SW_HIDE);
}

void TShapeWindow::HandleButton1Msg( UINT )
{ selected_shape=local_shape;
  Show(SW_HIDE);
}

void TShapeWindow::Paint(TDC &DC,BOOL, TRect &r1)
{ int bx=local_shape.firstx().x(),by=local_shape.firsty().y(),
      ex=local_shape.lastx().x(),ey=local_shape.lasty().y();
  int magnification= (r.right+1-r.left)/(ex+1-bx);
  if(magnification> (r.bottom+1-r.top)/(ey+1-by))
     magnification=(r.bottom+1-r.top)/(ey+1-by);
  if(magnification==0)magnification=1;
  if(magnification>25)magnification=25;
  char prevmsg[20];
  sprintf(prevmsg,"Preview:      Mag:%d",magnification);
  PreviewMsg->SetText(prevmsg);
  DC.MoveTo(r.left-1,r.top-1);
  DC.LineTo(r.left-1,r.bottom+1);
  DC.LineTo(r.right+1,r.bottom+1);
  DC.LineTo(r.right+1,r.top-1);
  DC.LineTo(r.left-1,r.top-1);
  DC.SelectObject(*Pens[SET]);
  DC.SelectObject(*Brushes[SET]);
  int xoffset=((r.right+1+r.left)-magnification*(ex+1-bx))/2;
  if(xoffset<r.left)xoffset=r.left;
  int yoffset=((r.bottom+1+r.top)-magnification*(ey+1-by))/2;
  if(yoffset<r.top)yoffset=r.top;
  int sz=magnification-(magnification>3);
  for(int iw=0;iw<local_shape.pop;iw++)
  { int x=xoffset+magnification*(local_shape.v[iw].x()-bx);
    int y=yoffset+magnification*(local_shape.v[iw].y()-by);
    if(magnification>1)DC.Rectangle(x,y,x+sz,y+sz);
    else DC.SetPixel(x,y,Colors[SET]);
  }
  TPen *pen=new TPen(TColor(100,100,100));
  DC.SelectObject(*pen);
  xoffset+=magnification/2-1-bx*magnification;
  yoffset+=magnification/2-1-by*magnification;
  DC.MoveTo(xoffset,r.top-1);  DC.LineTo(xoffset,r.bottom+1);
  DC.MoveTo(r.left-1,yoffset);  DC.LineTo(r.right-1,yoffset);
  delete pen;
}

void TShapeWindow::HandleButton3Msg( UINT )
{ TOpenSaveDialog::TData FileData(OFN_FILEMUSTEXIST | OFN_HIDEREADONLY |
			   OFN_PATHMUSTEXIST,
   "All Files (*.*)|*.*|Life Files (*.lfe)|*.lfe|",0,0,"lfe");
  strcpy(FileData.FileName,"");
  if((TFileOpenDialog(this,FileData)).Execute()==IDOK)
  {  cellpop temp=readfile(FileData.FileName);
     if(temp.error())return;
     local_shape=temp;
     selected_shape=local_shape;show();
  }
}

void TShapeWindow::HandleButton4Msg( UINT )
{ char InputText[6];static int lg=10;
  sprintf(InputText, "%d", lg);
  if ( GetApplication()->ExecDialog(
  new TInputDialog(this, "Input Size of Line of Cells:","", InputText,
     sizeof InputText))==IDOK )lg = atoi(InputText);
  local_shape=cellpop(lg);
  selected_shape=local_shape;show();
}

void TShapeWindow::HandleButton5Msg( UINT )
{ local_shape.transform(1);
  selected_shape=local_shape;show();
}

void TShapeWindow::HandleButton6Msg( UINT )
{ local_shape.transform(5);
  selected_shape=local_shape;show();
}

void TShapeWindow::HandleButton7Msg( UINT )
{ TOpenSaveDialog::TData FileData(OFN_FILEMUSTEXIST | OFN_HIDEREADONLY |
			   OFN_PATHMUSTEXIST,
   "All Files (*.*)|*.*|Life Libraries (*.lib)|*.lib|",0,"","lib");
  strcpy(FileData.FileName,"");
  if((TFileOpenDialog(this,FileData)).Execute()==IDOK)
  {  readshapelib(FileData.FileName);
     SetupList();
  }
}

DEFINE_RESPONSE_TABLE1( TShapeWindow, TWindow )
    EV_CHILD_NOTIFY_ALL_CODES( ID_LISTBOX, HandleListBoxMsg ),
    EV_CHILD_NOTIFY_ALL_CODES( ID_BUTTON1, HandleButton1Msg ),
    EV_CHILD_NOTIFY_ALL_CODES( ID_BUTTON2, HandleButton2Msg ),
    EV_CHILD_NOTIFY_ALL_CODES( ID_BUTTON3, HandleButton3Msg ),
    EV_CHILD_NOTIFY_ALL_CODES( ID_BUTTON4, HandleButton4Msg ),
    EV_CHILD_NOTIFY_ALL_CODES( ID_BUTTON5, HandleButton5Msg ),
    EV_CHILD_NOTIFY_ALL_CODES( ID_BUTTON6, HandleButton6Msg ),
    EV_CHILD_NOTIFY_ALL_CODES( ID_BUTTON7, HandleButton7Msg ),
END_RESPONSE_TABLE;
